@@ -110,6 +110,8 @@ extern int errno;
 #include "../gnulib/lib/error.h"
 #include "../gnulib/lib/human.h"
 #include "dirname.h"
+#include "closeout.h"
+#include "nextelem.h"
 
 /* Note that this evaluates C many times.  */
 #ifdef _LIBC
@@ -137,7 +139,7 @@ int check_existence = 0;
 /* What to separate the results with. */
 static int separator = '\n';
 
-char *next_element ();
+
 
 
 /* Read in a 16-bit int, high byte first (network byte order).  */
@@ -255,8 +257,8 @@ struct casefolder
 
 
 
-typedef int (*visitfunc)(const char *tested_filename,
-			 const char *printed_filename,
+typedef int (*visitfunc)(const char *munged_filename,
+			 const char *original_filename,
 			 void *context);
 
 struct visitor
@@ -271,14 +273,14 @@ static struct visitor *inspectors = NULL;
 static struct visitor *lastinspector = NULL;
 
 static int
-process_filename(const char *tested_filename, const char *printed_filename)
+process_filename(const char *munged_filename, const char *original_filename)
 {
   int result = VISIT_CONTINUE;
   const struct visitor *p = inspectors;
   
   while ( (VISIT_CONTINUE == result) && (NULL != p) )
     {
-      result = (p->inspector)(tested_filename, printed_filename, p->context);
+      result = (p->inspector)(munged_filename, original_filename, p->context);
       p = p->next;
     }
 
@@ -310,28 +312,28 @@ add_visitor(visitfunc fn, void *context)
 
 
 static int
-visit_justprint(const char *testname, const char *printname, void *context)
+visit_justprint(const char *munged_filename, const char *original_filename, void *context)
 {
   (void) context;
-  (void) testname;
-  fputs(printname, stdout);
+  (void) munged_filename;
+  fputs(original_filename, stdout);
   putchar(separator);
   return VISIT_CONTINUE;
 }
 
 static int
-visit_exists(const char *testname, const char *printname, void *context)
+visit_exists(const char *munged_filename, const char *original_filename, void *context)
 {
   struct stat st;
   (void) context;
-  (void) testname;
+  (void) munged_filename;
 
-  /* testname has been converted in some way (to lower case,
-   * or is just the basename of the file), and printname has not.  
-   * Hence only printname is still actually the name of the file 
+  /* munged_filename has been converted in some way (to lower case,
+   * or is just the basename of the file), and original_filename has not.  
+   * Hence only original_filename is still actually the name of the file 
    * whose existence we would need to check.
    */
-  if (stat(printname, &st) != 0)
+  if (stat(original_filename, &st) != 0)
     {
       return VISIT_REJECTED;
     }
@@ -342,30 +344,30 @@ visit_exists(const char *testname, const char *printname, void *context)
 }
 
 static int
-visit_substring_match_nocasefold(const char *testname, const char *printname, void *context)
+visit_substring_match_nocasefold(const char *munged_filename, const char *original_filename, void *context)
 {
   const char *pattern = context;
-  (void) printname;
+  (void) original_filename;
 
-  if (NULL != strstr(testname, pattern))
+  if (NULL != strstr(munged_filename, pattern))
     return VISIT_CONTINUE;
   else
     return VISIT_REJECTED;
 }
 
 static int
-visit_substring_match_casefold(const char *testname, const char *printname, void *context)
+visit_substring_match_casefold(const char *munged_filename, const char *original_filename, void *context)
 {
   struct casefolder * p = context;
-  size_t len = strlen(testname);
+  size_t len = strlen(munged_filename);
 
-  (void) printname;
+  (void) original_filename;
   if (len+1 > p->buffersize)
     {
       p->buffer = xrealloc(p->buffer, len+1); /* XXX: consider using extendbuf(). */
       p->buffersize = len+1;
     }
-  lc_strcpy(p->buffer, testname);
+  lc_strcpy(p->buffer, munged_filename);
   
   
   if (NULL != strstr(p->buffer, p->pattern))
@@ -376,11 +378,11 @@ visit_substring_match_casefold(const char *testname, const char *printname, void
 
 
 static int
-visit_globmatch_nofold(const char *testname, const char *printname, void *context)
+visit_globmatch_nofold(const char *munged_filename, const char *original_filename, void *context)
 {
   const char *glob = context;
-  (void) printname;
-  if (fnmatch(glob, testname, 0) != 0)
+  (void) original_filename;
+  if (fnmatch(glob, munged_filename, 0) != 0)
     return VISIT_REJECTED;
   else
     return VISIT_CONTINUE;
@@ -388,21 +390,21 @@ visit_globmatch_nofold(const char *testname, const char *printname, void *contex
 
 
 static int
-visit_globmatch_casefold(const char *testname, const char *printname, void *context)
+visit_globmatch_casefold(const char *munged_filename, const char *original_filename, void *context)
 {
   const char *glob = context;
-  (void) printname;
-  if (fnmatch(glob, testname, FNM_CASEFOLD) != 0)
+  (void) original_filename;
+  if (fnmatch(glob, munged_filename, FNM_CASEFOLD) != 0)
     return VISIT_REJECTED;
   else
     return VISIT_CONTINUE;
 }
 
 static int
-visit_stats(const char *testname, const char *printname, void *context)
+visit_stats(const char *munged_filename, const char *original_filename, void *context)
 {
   struct locate_stats *p = context;
-  size_t len = strlen(printname);
+  size_t len = strlen(original_filename);
   const char *s;
   int highbit, whitespace, newline;
 
@@ -410,7 +412,7 @@ visit_stats(const char *testname, const char *printname, void *context)
   p->total_filename_length += len;
   
   highbit = whitespace = newline = 0;
-  for (s=printname; *s; ++s)
+  for (s=original_filename; *s; ++s)
     {
       if ( (int)(*s) & 128 )
 	highbit = 1;
@@ -514,7 +516,7 @@ new_locate (char *pathpart,
 	}
 
       /* We add visit_exists() as late as possible to reduce the
-       * number of stat() calls.  (Idea by Bas van Gompel).
+       * number of stat() calls.
        */
       if (check_existence)
 	add_visitor(visit_exists, NULL);
@@ -697,33 +699,13 @@ usage (stream)
 {
   fprintf (stream, _("\
 Usage: %s [-d path | --database=path] [-e | --existing]\n\
-      [-i | --ignore-case] [--wholepath] [--basename] [--limit=N | -l N]\n\
+      [-i | --ignore-case] [-w | --wholename] [-b | --basename] \n\
+      [--limit=N | -l N] [-S | --statistics] [-0 | --null] [-c | --count]\n\
       [--version] [--help] pattern...\n"),
 	   program_name);
   fputs (_("\nReport bugs to <bug-findutils@gnu.org>.\n"), stream);
 }
 
-static void
-sanity_check_dbpath(const char *path)
-{
-  size_t len;
-  
-  if (':' == path[0])
-    {
-      error(0, 0, _("warning: locate database path `%s' contains a leading colon, which is not a valid database name"), path);
-    }
-
-  len = strlen(path);
-  if (len > 0u)
-    {
-      if (':' == path[len-1u])
-	{
-	  error(0, 0, _("warning: locate database path `%s' contains a trailing colon, which is not a valid database name"), path);
-	}
-    }
-}
-
-
 static struct option const longopts[] =
 {
   {"database", required_argument, NULL, 'd'},
@@ -734,6 +716,7 @@ static struct option const longopts[] =
   {"null", no_argument, NULL, '0'},
   {"count", no_argument, NULL, 'c'},
   {"wholename", no_argument, NULL, 'w'},
+  {"wholepath", no_argument, NULL, 'w'}, /* Synonym. */
   {"basename", no_argument, NULL, 'b'},
   {"stdio", no_argument, NULL, 's'},
   {"mmap",  no_argument, NULL, 'm'},
@@ -765,6 +748,7 @@ main (argc, argv)
 #endif
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
+  atexit (close_stdout);
 
   dbpath = getenv ("LOCATE_PATH");
   if (dbpath == NULL)
@@ -772,7 +756,7 @@ main (argc, argv)
 
   check_existence = 0;
 
-  while ((optc = getopt_long (argc, argv, "bcd:eil:sm0S", longopts, (int *) 0)) != -1)
+  while ((optc = getopt_long (argc, argv, "bcd:eil:sm0Sw", longopts, (int *) 0)) != -1)
     switch (optc)
       {
       case '0':
@@ -854,15 +838,13 @@ main (argc, argv)
 	}
     }
   
-  sanity_check_dbpath(dbpath);
-
   for (; stats || optind < argc; optind++)
     {
       char *e;
       const char *needle;
-      next_element (dbpath);	/* Initialize.  */
+      next_element (dbpath, 0);	/* Initialize.  */
       needle = stats ? NULL : argv[optind];
-      while ((e = next_element ((char *) NULL)) != NULL)
+      while ((e = next_element ((char *) NULL, 0)) != NULL)
 	{
 	  statistics.compressed_bytes = 
 	    statistics.total_filename_count = 
@@ -870,6 +852,15 @@ main (argc, argv)
 	    statistics.whitespace_count = 
 	    statistics.newline_count = 
 	    statistics.highbit_filename_count = 0u;
+
+	  if (0 == strlen(e) || 0 == strcmp(e, "."))
+	    {
+	      /* Use the default database name instead (note: we
+	       * don't use 'dbpath' since that might itself contain a 
+	       * colon-separated list.
+	       */
+	      e = LOCATE_DB;
+	    }
 	  
 	  found += new_locate (needle, e, ignore_case, print, basename_only, use_limit, limit, stats);
 	}
